/**CFile***********************************************************************

  FileName    [compileBEval.c]

  PackageName [compile]

  Synopsis    []

  Description [Conversion from scalar expressions to boolean expressions.]

  SeeAlso     [optional]

  Author      [Alessandro Cimatti and Marco Roveri]

  Copyright   [
  This file is part of the ``compile'' package of NuSMV version 2.
  Copyright (C) 2000-2001 by ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include "compile/compile.h"
#include "compileInt.h"
#include "utils/assoc.h"
#include "parser/symbols.h"
#include "parser/mbp_symbols.h"
#include "utils/error.h"
#include "utils/utils_io.h" /* for indent_node */
#include "utils/range.h"
#include "enc/enc.h"

static char rcsid[] UTIL_UNUSED = "$Id: compileBEval.c,v 1.13.4.12.2.1.2.18.4.1 2006/08/23 16:02:16 nusmv Exp $";

/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/


static hash_ptr expr2bexpr_hash = (hash_ptr) NULL;

void init_expr2bexp_hash() 
{
  nusmv_assert(expr2bexpr_hash == (hash_ptr) NULL);
  expr2bexpr_hash = new_assoc();
}

void clear_expr2bexp_hash()
{
  if (expr2bexpr_hash != (hash_ptr) NULL) {
    free_assoc(expr2bexpr_hash);
    expr2bexpr_hash = (hash_ptr) NULL;
  }
}

static node_ptr make_key(node_ptr expr, boolean a, boolean b) 
{
  int j = 0;  

  j += a? 1 : 0;
  j += b? 2 : 0;

  return find_node(CONTEXT, expr, (node_ptr) j);
}

static void expr2bexpr_hash_insert_entry(node_ptr expr, node_ptr bexpr, 
					 boolean a, boolean b)
{
  nusmv_assert(expr2bexpr_hash != (hash_ptr) NULL);
  insert_assoc(expr2bexpr_hash, make_key(expr, a, b), bexpr);
}

static node_ptr 
expr2bexpr_hash_lookup_entry(node_ptr expr, boolean a, boolean b)
{
  nusmv_assert(expr2bexpr_hash != (hash_ptr) NULL);
  return find_assoc(expr2bexpr_hash, make_key(expr, a, b));
}


/**AutomaticStart*************************************************************/

/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/
 
static node_ptr 
expr2bexpr_recur ARGS((BddEnc_ptr enc, SymbLayer_ptr,
		       node_ptr, node_ptr, boolean));
static node_ptr 
scalar_atom2bexpr ARGS((BddEnc_ptr, SymbLayer_ptr,
			node_ptr, node_ptr, boolean));


/**AutomaticEnd***************************************************************/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [Converts a scalar expression into a boolean expression.]

  Description        [Takes an scalar expression intended to evaluate
  to boolean, maps through booleans down to the atomic scalar
  propositions, builds the corresponding boolean function, and returns
  the resulting boolean expression.

  The conversion of atomic scalar proposition is currently performed
  by generating the corresponding ADD, and then printing it in terms
  of binary variables.

  If one or more determinization variable must be created
  (i.e. non-determinism must be allowed) then det_layer is the
  SymbLayer instance to be filled with the newly created
  determinization variables. If non-determinism is not allowed, specify 
  NULL as det_layer value. In this case you can use detexpr2bexpr as well.

  The input expression will be processed with Nil context (for
  flattened expr this does not matter).]

  SideEffects        [None]

  SeeAlso            [Compile_detexpr2bexpr, expr2bexpr_recur]

******************************************************************************/
Expr_ptr Compile_expr2bexpr(BddEnc_ptr enc, 
			    SymbLayer_ptr det_layer, 
			    Expr_ptr expr)
{
  node_ptr res;
  int temp = yylineno;

  if (expr == EXPR(NULL)) return expr;

  yylineno = node_get_lineno( NODE_PTR(expr) );
  res = expr2bexpr_recur(enc, det_layer, expr, Nil, false);
  yylineno = temp;
  return EXPR(res);
}


/**Function********************************************************************

  Synopsis           [Converts a scalar expression into a boolean expression.]

  Description        [Takes an scalar expression intended to evaluate
  to boolean, maps through booleans down to the atomic scalar
  propositions, builds the corresponding boolean function, and returns
  the resulting boolean expression.

  The conversion of atomic scalar proposition is currently performed
  by generating the corresponding ADD, and then printing it in terms
  of binary variables.

  An error is returned if determinization variables are introduced in
  the booleanization process.

  The input expression will be processed with Nil context (for
  flattened expr this does not matter).]

  SideEffects        [None]

  SeeAlso            [Compile_expr2bexpr, expr2bexpr_recur]

******************************************************************************/
Expr_ptr Compile_detexpr2bexpr(BddEnc_ptr enc, Expr_ptr expr)
{
  node_ptr res;
  int temp = yylineno;

  if (expr == EXPR(NULL)) return expr;
  
  yylineno = node_get_lineno(expr);
  res = expr2bexpr_recur(enc, SYMB_LAYER(NULL), expr, Nil, false);

  yylineno = temp;
  return EXPR(res);
}


/*---------------------------------------------------------------------------*/
/* Definition of internal functions                                          */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of static functions                                            */
/*---------------------------------------------------------------------------*/


/**Function********************************************************************

  Synopsis           [Converts a generic expression into a boolean expression.]

  Description        [Takes an expression intended to evaluate to boolean,
   maps through booleans down to the atomic scalar propositions,
   builds the corresponding boolean function,
   and returns the resulting boolean expression.

   The conversion of atomic scalar proposition is currently performed
   by generating the corresponding ADD, and then printing it in terms
   of binary variables.

   The introduction of determinization variables is allowed only if flag
   <tt>allow_nondet</tt> is set to true.

   The input expression may be normal (not flattened), flattened or
   expanded. Parameter 'context' is used if the expression is not flattened.]

  SideEffects        [None]

  SeeAlso            [Compile_expr2bexpr, detexpr2bexpr]

******************************************************************************/
static node_ptr expr2bexpr_recur(BddEnc_ptr enc, 
				 SymbLayer_ptr det_layer, 
				 node_ptr expr,
				 node_ptr context,
				 boolean in_next)
{
  SymbTable_ptr symb_table = BaseEnc_get_symbol_table(BASE_ENC(enc));
  node_ptr res; 

  if (expr == Nil) return Nil;

  res = expr2bexpr_hash_lookup_entry(expr, in_next, 
				     det_layer != SYMB_LAYER(NULL));
  if (res != (node_ptr) NULL) return res;

  switch (node_get_type(expr)) {

  case FAILURE:
  case TRUEEXP:
  case FALSEEXP:
    res = expr;
    break;

  case NUMBER: 
    if (((int)car(expr) | 1) != 1) { /* not boolean value */
      rpterr("Number can not be casted to boolean (except 0 and 1)");
    }
    res = (int)car(expr) 
      ? find_node(TRUEEXP, Nil, Nil) 
      : find_node(FALSEEXP, Nil, Nil);
    break;
    
  case NUMBER_WORD: 
  case NUMBER_FRAC: 
  case NUMBER_REAL:
  case NUMBER_EXP:
    rpterr("Word or real constant cannot be casted to boolean");

  case BIT:
    {
      node_ptr name = CompileFlatten_resolve_name(symb_table, expr, context);
      nusmv_assert(SymbTable_is_symbol_var(symb_table, name)); 
      
      res = name;
      break;
    }

  case DOT:
  case ATOM:
  case ARRAY: 
    {
      node_ptr name = CompileFlatten_resolve_name(symb_table, expr, context);
      node_ptr param = lookup_param_hash(name);

      /* no check for ambiguity is required, since type checker did it */

      /* parameter */
      if (param != (node_ptr)NULL) {
        /* process the value of the parameter */
	res = expr2bexpr_recur(enc, det_layer, param, context, in_next);
        break;
      }

      /* define */
      if (SymbTable_is_symbol_define(symb_table, name)) {
	/* define (rhs must be boolean, recur to check) */
	node_ptr body = SymbTable_get_define_flatten_body(symb_table, name);
	if (body == (node_ptr) NULL) error_undefined(name); 
	
	res = expr2bexpr_recur(enc, det_layer, body, context, in_next);
      }
      /* variable */
      else if (SymbTable_is_symbol_var(symb_table, name)) {
	/* variable, must be boolean */
	if (!SymbType_is_boolean_enum(
		      SymbTable_get_var_type(symb_table, name))) {
	  rpterr("Unexpected non boolean variable");
	} 
	    
	if (in_next) res = find_node(NEXT, name, Nil);
	else res = name;
      }

      else {
	/* unknow identifier. This code should be impossible since the 
	   type checker already checked the expr for being correct
	*/
	print_sexp(nusmv_stderr, expr);
	rpterr("Unexpected data structure"); 
      }
    }
    break;
    
  case CAST_BOOL:
    return scalar_atom2bexpr(enc, det_layer, expr, context, in_next);

  case CAST_WORD1:
    rpterr("Word-1 cannot be casted to boolean");

  case LSHIFT:
  case RSHIFT:
  case LROTATE:
  case RROTATE:
    rpterr("Binary word expression cannot be casted to boolean");
    
  case CONCATENATION:
  case BIT_SELECTION:
    rpterr("Word expression cannot be casted to boolean");
     
  case SIGN_EXTEND:
    rpterr("Sign extend cannot be casted to boolean");
     
  case NOT:
    res = find_node(NOT, expr2bexpr_recur(enc, det_layer,
					  car(expr), context, in_next),
		    Nil);
    break;

  case UMINUS:
    res = find_node(UMINUS, expr2bexpr_recur(enc, det_layer,
					  car(expr), context, in_next),
		    Nil);
    break;
  
  case CONS:
  case AND:
  case OR:
  case XOR:
  case XNOR:
  case IMPLIES:
  case IFF:
    res = find_node(node_get_type(expr),
		    expr2bexpr_recur(enc, det_layer, 
				     car(expr), context, in_next),
		    expr2bexpr_recur(enc, det_layer,
				     cdr(expr), context, in_next));

    break;

  case CASE:
    res = find_node(CASE,
        	    find_node(COLON,
			      expr2bexpr_recur(enc, det_layer, car(car(expr)),
					       context, in_next),
			      expr2bexpr_recur(enc, det_layer, cdr(car(expr)),
					       context, in_next)),
		    expr2bexpr_recur(enc, det_layer, 
				     cdr(expr), context, in_next));
    break;
    
  case NEXT: 
    nusmv_assert(!in_next);
    res = expr2bexpr_recur(enc, det_layer, car(expr), context, true);
    break;
    
  case EQDEF:
    {
      node_ptr var_name;
      node_ptr var;
      node_ptr lhs = car(expr);
      node_ptr rhs = cdr(expr);

      switch (node_get_type(lhs)) {
      case SMALLINIT:
        var_name = car(lhs);
        var = car(lhs);
	break;
	
      case NEXT:
        var_name = car(lhs);
        var = lhs;
	break;
	
      default:
        var_name = lhs;
        var = lhs;
      }

      {
        node_ptr name = CompileFlatten_resolve_name(symb_table, var_name, 
						    context);

        if (!SymbTable_is_symbol_declared(symb_table, name)) {
	  error_undefined(name);
        } 
	
	if (SymbTable_is_symbol_var(symb_table, name)) {
	  if (SymbType_is_boolean_enum(
		  SymbTable_get_var_type(symb_table, name))) {
            /* boolean variable, rhs will be determinized */
            if (node_get_type(lhs) == NEXT) var = find_node(NEXT, name, Nil);
            else var = name;
            
            res = find_node(IFF, var, 
		      expr2bexpr_recur(enc, det_layer, rhs, context, in_next));
          }
	  else {
            /* scalar variable */
            nusmv_assert(!in_next);
            res = scalar_atom2bexpr(enc, det_layer, expr, context, in_next);
          }
	} 
	else { rpterr("Unexpected data structure, variable was expected"); }
      }
      break;
    } /* end of case EQDEF */

    /* Predicates over scalar terms (guaranteed to return boolean) */
  case LT:
  case GT:
  case LE:
  case GE:

    /* Signed predicates over scalar terms (guaranteed to return boolean) */
  case SLT:
  case SGT:
  case SLE:
  case SGE:

  /* Predicates over possibly scalar terms (guaranteed to return boolean) */
  case EQUAL:
  case NOTEQUAL:
  
  /* Function symbols over scalar terms Might return boolean, but
    check is needed Assumption: if boolean is returned, then the
    function is determinized by introducing a variable on the {0,1}
    leaves. */
  case PLUS:
  case MINUS:
  case TIMES:
  case DIVIDE:
  case MOD:
  case UNION:
  case SETIN: 
    res = scalar_atom2bexpr(enc, det_layer, expr, context, in_next);
    break;
  

  /* UNARY CTL/LTL OPERATORS */
  case EX:
  case AX:
  case EG:
  case AG:
  case EF:
  case AF:
  case OP_NEXT:
  case OP_PREC:
  case OP_NOTPRECNOT:
  case OP_FUTURE:
  case OP_ONCE:
  case OP_GLOBAL:
  case OP_HISTORICAL:
    res = find_node(node_get_type(expr),
		  expr2bexpr_recur(enc, det_layer, car(expr), context, in_next),
		    Nil);
    break;
  

  /* BINARY CTL/LTL OPERATORS */
  case EU:
  case AU:
  case MINU:
  case MAXU:
  case UNTIL:
  case RELEASES:
  case SINCE: 
  case TRIGGERED: 
    res = find_node(node_get_type(expr),
	       expr2bexpr_recur(enc, det_layer, car(expr), context, in_next),
	       expr2bexpr_recur(enc, det_layer, cdr(expr), context, in_next));
    break;
    

  /* BOUNDED TEMPORAL OPERATORS (cdr is range, no recursion needed) */
  case EBF:
  case ABF:
  case EBG:
  case ABG:
  case EBU:
  case ABU: 
    res = find_node(node_get_type(expr),
		   expr2bexpr_recur(enc, det_layer, car(expr), context, in_next),
		    cdr(expr));
    break;
  

  case CONTEXT:
    res = expr2bexpr_recur(enc, det_layer, cdr(expr), car(expr), in_next);
    break;
  

  case TWODOTS: 
    rpterr("Unexpected TWODOTS node");
    res = (node_ptr) NULL;
    break;

#if HAVE_MBP  /* MBP specific expressions. These expressions are converted
		 to boolean only when a boolean module is printed out */
  case MBP_SPEC_WRAPPER: /* skip the player name (left) and process
			    the expression (right child) */
    res = find_node(MBP_SPEC_WRAPPER,
		    car(expr),
	        expr2bexpr_recur(enc, det_layer, cdr(expr), context, in_next));
		    
    break;
		
    /* contain one or two lists. Process them the same way since Nil
       will be skipped */
  case MBP_EXP_LIST:
  case MBP_TWO_EXP_LISTS:
    res = find_node(node_get_type(expr),
		    expr2bexpr_recur(enc, det_layer, 
				     car(expr), context, in_next),
		    expr2bexpr_recur(enc, det_layer,
				     cdr(expr), context, in_next));

    break;
#endif /* HAVE_MBP */

  default: 
    internal_error("expr2bexpr_recur: type = %d\n", node_get_type(expr));
    res = (node_ptr) NULL;
  }

  /* updates the results hash  */
  if (res != (node_ptr) NULL) {
    expr2bexpr_hash_insert_entry(expr, res, in_next, 
				 det_layer != SYMB_LAYER(NULL));
  }

  return res;
}


/**Function********************************************************************

  Synopsis [Converts an atomic expression into the corresponding
  (boolean) expression.]

  Description        [Takes an atomic expression and converts it into
  a corresponding boolean expression.

  The introduction of determinization variables is allowed only if the layer 
  <tt>det_layer</tt> is not NULL]

  SideEffects        [A new boolean variable might be introduced.]

  SeeAlso            []

******************************************************************************/
static node_ptr scalar_atom2bexpr(BddEnc_ptr enc, SymbLayer_ptr det_layer, 
				  node_ptr expr, node_ptr context,
				  boolean in_next)
{
  node_ptr res;
  int temp = yylineno;
  add_ptr bool_add = BddEnc_expr_to_add(enc, 
				(in_next) ? find_node(NEXT, expr, Nil) : expr, 
				context);

  yylineno = node_get_lineno(expr);
  res = BddEnc_add_to_expr(enc, bool_add, det_layer);
  add_free(BddEnc_get_dd_manager(enc), bool_add);
  yylineno = temp;

  return res;
}




