/**CFile***********************************************************************

  FileName    [operators.c]

  PackageName [enc]

  Synopsis    [These operators are used by dd package]

  Description [Functions like add_plus, add_equal, etc., call these operators]

  SeeAlso     [operators.h]

  Author      [Marco Roveri, Roberto Cavada]

  Copyright   [
  This file is part of the ``enc'' package of NuSMV version 2. 
  Copyright (C) 2003 by ITC-irst.

  NuSMV version 2 is free software; you can redistribute it and/or 
  modify it under the terms of the GNU Lesser General Public 
  License as published by the Free Software Foundation; either 
  version 2 of the License, or (at your option) any later version.

  NuSMV version 2 is distributed in the hope that it will be useful, 
  but WITHOUT ANY WARRANTY; without even the implied warranty of 
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public 
  License along with this library; if not, write to the Free Software 
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA.

  For more information on NuSMV see <http://nusmv.irst.itc.it>
  or email to <nusmv-users@irst.itc.it>.
  Please report bugs to <nusmv-users@irst.itc.it>.

  To contact the NuSMV development board, email to <nusmv@irst.itc.it>. ]

******************************************************************************/

#include "operators.h"
#include "parser/symbols.h"
#include "utils/error.h"
#include "utils/ustring.h"
#include "opt/opt.h"
#include "utils/WordNumber.h"

static char rcsid[] UTIL_UNUSED = "$Id: operators.c,v 1.1.2.5.4.17.4.1 2006/08/23 16:02:16 nusmv Exp $";


/*---------------------------------------------------------------------------*/
/* Constant declarations                                                     */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Type declarations                                                         */
/*---------------------------------------------------------------------------*/

typedef int (*INTPFII)(int, int);


/*---------------------------------------------------------------------------*/
/* macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/* Variable declarations                                                     */
/*---------------------------------------------------------------------------*/
EXTERN int yylineno;
EXTERN options_ptr options;
/*---------------------------------------------------------------------------*/
/* Static function prototypes                                                */
/*---------------------------------------------------------------------------*/


/*---------------------------------------------------------------------------*/
/* Definition of exported functions                                          */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis           [ performs logical or bitwise AND on two nodes.]

  Description        [ Nodes can be integers with values 0 and 1 (logical AND) 
  or Word constants (bitwise AND). All other combinations are illegal.]

  SideEffects        []

******************************************************************************/
node_ptr node_and(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return  find_node(NUMBER, (node_ptr)((int)car(n1) & (int)car(n2)), Nil);
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_and(WORD_NUMBER(car(n1)),
					      WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("&", n1, n2);

  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [ performs logical or bitwise OR on two nodes.]

  Description        [ Nodes can be integers with values 0 and 1 (logical OR) 
  or Word constants (bitwise OR). All other combinations are illegal.]

  SideEffects        []

******************************************************************************/
node_ptr node_or(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return  find_node(NUMBER, (node_ptr)((int)car(n1) | (int)car(n2)), Nil);
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_or(WORD_NUMBER(car(n1)),
					     WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("|", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [ performs logical or bitwise NOT on a node.]

  Description        [ Node can be an integer with values 0 or 1 (logical NOR) 
  or a Word constant (bitwise OR). All other combinations are illegal.

  NOTE: At the momement, CUDD does not have unary 'apply', so
  you have to write a unary operator in the form of a binary one which
  actually applies to the first operand only]

  SideEffects        []

******************************************************************************/
node_ptr node_not(node_ptr n, node_ptr this_node_is_not_used)
{ 
  if (node_get_type(n) == FAILURE) return n; /* error in previous expr */

  if (node_get_type(n) == NUMBER) {
    return (int)car(n) ? zero_number : one_number;
  }
  else if(node_get_type(n) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_not(WORD_NUMBER(car(n))), Nil);
  }
  else error_not_proper_number("", n);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [ performs logical or bitwise IFF on two nodes.]

  Description        [ Nodes can be integers with values 0 and 1 (logical IFF) 
  or Word constants (bitwise IFF). All other combinations are illegal.]

  SideEffects        []

******************************************************************************/
node_ptr node_iff(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return  find_node(NUMBER, (node_ptr)((int)car(n1) == (int)car(n2)), Nil);
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_iff(WORD_NUMBER(car(n1)),
					      WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("iff", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [ performs logical or bitwise XOR on two nodes.]

  Description        [ Nodes can be integers with values 0 and 1 (logical XOR) 
  or Word constants (bitwise XOR). All other combinations are illegal.]

  SideEffects        []

******************************************************************************/
node_ptr node_xor(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return  find_node(NUMBER, (node_ptr)((int)car(n1) ^ (int)car(n2)), Nil);
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_xor(WORD_NUMBER(car(n1)),
					      WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("xor", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis        [ performs logical or bitwise IMPLIES on two nodes.]

  Description     [Nodes can be integers with values 0 and 1 (logical IMPLIES) 
  or Word constants (bitwise IMPLIES). All other combinations are illegal.]

  SideEffects     []

******************************************************************************/
node_ptr node_implies(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return  find_node(NUMBER, (node_ptr)((!(int)car(n1)) | (int)car(n2)), Nil);
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_implies(WORD_NUMBER(car(n1)),
						 WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("implies", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 (symbol one_number) if 
  the nodes are the same, and value 0 (symbol zero_number) otherwise]

  Description        [
  In NuSMV an constant is equal to another constant then this 
  constants are actually the same and representable by the same node.
  The only exception is if one operand is boolean and the other 
  is Word -- they also may be equal]

  SideEffects        []

  SeeAlso            [node_setin]

******************************************************************************/
node_ptr node_equal(node_ptr n1, node_ptr n2)
{
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if(n1 == n2) return(one_number);

  /* additional equality test: Word[1] * boolean. 
     This test is required only if any Word expression is represented as one ADD
  */
  /* n1 is boolean and n2 is Word[1] and they are equal. */
  if (NUMBER == node_get_type(n1) && NUMBER_WORD == node_get_type(n2) &&
      1 == WordNumber_get_width(WORD_NUMBER(car(n2))) &&
      (int)car(n1) == WordNumber_get_bit(WORD_NUMBER(car(n2)), 0)) {
    return(one_number);
  }

  /* n1 is Word[1] and n2 is boolean and they are equal */
  if (NUMBER == node_get_type(n2) && NUMBER_WORD == node_get_type(n1) &&
      1 == WordNumber_get_width(WORD_NUMBER(car(n1))) &&
      (int)car(n2) == WordNumber_get_bit(WORD_NUMBER(car(n1)), 0)) {
    return(one_number);
  }

  return(zero_number);
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 (symbol one_number) if 
  the nodes are of different values, and value 0 (symbol zero_number) otherwise]

  Description        [
  In NuSMV an constant is equal to another constant then this 
  constants are actually the same and representable by the same node.
  The only exception is if one operand is boolean and the other 
  is Word -- they also may be equal]

  SideEffects        []

  SeeAlso            [node_setin]

******************************************************************************/
node_ptr node_not_equal(node_ptr n1, node_ptr n2)
{
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if(n1 == n2) return(zero_number);

  /* additional equality test: Word[1] * boolean.
     This test is required only if any Word expression is represented as one ADD
  */
  /* n1 is boolean and n2 is Word[1] and they are equal */
  if (NUMBER == node_get_type(n1) && NUMBER_WORD == node_get_type(n2) &&
      1 == WordNumber_get_width(WORD_NUMBER(car(n2))) &&
      (int)car(n1) == WordNumber_get_bit(WORD_NUMBER(car(n2)), 0)) {
    return(zero_number);
  }

  /* n1 is Word[1] and n2 is boolean and they are equal */
  if (NUMBER == node_get_type(n2) && NUMBER_WORD == node_get_type(n1) &&
      1 == WordNumber_get_width(WORD_NUMBER(car(n1))) &&
      (int)car(n2) == WordNumber_get_bit(WORD_NUMBER(car(n1)), 0)) {
    return(zero_number);
  }

  return(one_number);
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 if 
  the first node is less than the second one, and 0 - otherwise.]

  Description        [Nodes should be both NUMBER or both NUMBER_WORD]

  SideEffects        []

******************************************************************************/
node_ptr node_lt(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return ((int)car(n1) < (int)car(n2)) ? one_number : zero_number;
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return WordNumber_less(WORD_NUMBER(car(n1)), WORD_NUMBER(car(n2)))
      ? one_number : zero_number;
  }
  else error_not_proper_numbers("<", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 if 
  the first node is greater than the second one, and 0 - otherwise.]

  Description        [Nodes should be both NUMBER or both NUMBER_WORD]

  SideEffects        []

******************************************************************************/
node_ptr node_gt(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return (int)car(n1) > (int)car(n2) ? one_number : zero_number;
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return WordNumber_greater(WORD_NUMBER(car(n1)), WORD_NUMBER(car(n2)))
      ? one_number : zero_number;
  }
  else error_not_proper_numbers(">", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 if 
  the first node is less or equal than the second one, and 0 - otherwise.]

  Description        [Nodes should be both NUMBER or both NUMBER_WORD]

  SideEffects        []

******************************************************************************/
node_ptr node_le(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return ((int)car(n1) <= (int)car(n2)) ? one_number : zero_number;
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return WordNumber_less_or_equal(WORD_NUMBER(car(n1)), WORD_NUMBER(car(n2)))
      ? one_number : zero_number;
  }
  else error_not_proper_numbers("<=", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 if 
  the first node is greater or equal than the second one, and 0 - otherwise.]

  Description        [Nodes should be both NUMBER or both NUMBER_WORD]

  SideEffects        []

******************************************************************************/
node_ptr node_ge(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return (int)car(n1) >= (int)car(n2) ? one_number : zero_number;
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return WordNumber_greater_or_equal(WORD_NUMBER(car(n1)),
				       WORD_NUMBER(car(n2)))
      ? one_number : zero_number;
  }
  else error_not_proper_numbers(">=", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 if 
  the first node is signed less than the second one, and 0 - otherwise.]

  Description        [Nodes should be both NUMBER_WORD]

  SideEffects        []

******************************************************************************/
node_ptr node_slt(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return WordNumber_signed_less(WORD_NUMBER(car(n1)), WORD_NUMBER(car(n2)))
      ? one_number : zero_number;
  }
  else error_not_proper_numbers("signed <", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 if 
  the first node is signed greater than the second one, and 0 - otherwise.]

  Description        [Nodes should be both NUMBER_WORD]

  SideEffects        []

******************************************************************************/
node_ptr node_sgt(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return WordNumber_signed_greater(WORD_NUMBER(car(n1)), WORD_NUMBER(car(n2)))
      ? one_number : zero_number;
  }
  else error_not_proper_numbers("signed >", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 if 
  the first node is signed less or equal than the second one, and 0 - otherwise.]

  Description        [Nodes should be  both NUMBER_WORD]

  SideEffects        []

******************************************************************************/
node_ptr node_sle(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return WordNumber_signed_less_or_equal(WORD_NUMBER(car(n1)),
					   WORD_NUMBER(car(n2)))
      ? one_number : zero_number;
  }
  else error_not_proper_numbers("signed <=", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [returns NUMBER with value 1 if 
  the first node is signed greater or equal than the second one,
  and 0 - otherwise.]

  Description        [Nodes should be both NUMBER_WORD]

  SideEffects        []

******************************************************************************/
node_ptr node_sge(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return WordNumber_signed_greater_or_equal(WORD_NUMBER(car(n1)),
					      WORD_NUMBER(car(n2)))
      ? one_number : zero_number;
  }
  else error_not_proper_numbers("signed >=", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Negates the operand (unary minus)]

  Description        [Left node can be NUMBER or NUMBER_WORD, and the
  right one is Nil.]

  SideEffects        []

******************************************************************************/
node_ptr node_unary_minus(node_ptr n, node_ptr this_node_is_not_used)
{ 
  if (node_get_type(n) == FAILURE) return n; /* error in previous expr */

  if (node_get_type(n) == NUMBER) {
    return find_node(NUMBER, (node_ptr)( - (int)car(n)), Nil);
  }
  else if(node_get_type(n) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_unary_minus(WORD_NUMBER(car(n))), Nil);
  }
  else error_not_proper_number("- (unary)", n);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Adds two nodes]

  Description        [Nodes can be both NUMBER or both NUMBER_WORD.]

  SideEffects        []

******************************************************************************/
node_ptr node_plus(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return  find_node(NUMBER, (node_ptr)((int)car(n1) + (int)car(n2)), Nil);
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_plus(WORD_NUMBER(car(n1)),
					       WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("+", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Subtract two nodes]

  Description        [Nodes can be both NUMBER or both NUMBER_WORD.]

  SideEffects        []

******************************************************************************/
node_ptr node_minus(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return  find_node(NUMBER, (node_ptr)((int)car(n1) - (int)car(n2)), Nil);
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_minus(WORD_NUMBER(car(n1)),
						WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("-", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Multiplies two nodes]

  Description        [Nodes can be both NUMBER or both NUMBER_WORD.]

  SideEffects        []

******************************************************************************/
node_ptr node_times(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return  find_node(NUMBER, (node_ptr)((int)car(n1) * (int)car(n2)), Nil);
  }
  else if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_times(WORD_NUMBER(car(n1)),
						WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("*", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Divides two nodes]

  Description        [Nodes can be both NUMBER or both NUMBER_WORD.]

  SideEffects        []

******************************************************************************/
node_ptr node_divide(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  /* check the second operand for being zero */
  if (  (node_get_type(n2) == NUMBER && 0 == (int)car(n2)) 
     || (node_get_type(n2) == NUMBER_WORD && 
	 0L == WordNumber_get_value(WORD_NUMBER(car(n2)))) ) { 
    return failure_make("Division by zero", FAILURE_DIV_BY_ZERO, yylineno);
  }
 
  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    /* here is the check whether usual ANSI semantics of division is used */
    if (opt_use_ansi_c_div_op(options)) {
      return find_node(NUMBER, (node_ptr)((int)car(n1)/(int)car(n2)), Nil);
    }
    /* the semantics of the division is from old version of NuSMV */
    else {
      int a = (int) car(n1);
      int b = (int) car(n2);
      int r = a % b;
      int result = a/b - (r < 0);/*IF r < 0 THEN a/b - 1 ELSE a/b */
      return find_node(NUMBER, (node_ptr)result, Nil);
    }
  }
  
  if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_divide(WORD_NUMBER(car(n1)),
						 WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("/", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Computes the remainder of division of two nodes]

  Description        [Nodes can be both NUMBER or both NUMBER_WORD.]

  SideEffects        []

******************************************************************************/
node_ptr node_mod(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  /* check the second operand for being zero */
  if (  (node_get_type(n2) == NUMBER && 0 == (int)car(n2)) 
     || (node_get_type(n2) == NUMBER_WORD && 
	 0ULL == WordNumber_get_value(WORD_NUMBER(car(n2)))) ) { 
    return failure_make("Division by zero", FAILURE_DIV_BY_ZERO, yylineno);
  }

  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    /* here is the check whether usual ANSI semantics of division is used */
    if (opt_use_ansi_c_div_op(options)) {
      return find_node(NUMBER, (node_ptr)((int)car(n1) % (int)car(n2)), Nil);
    }
    /* the semantics of the division is from old version of NuSMV */
    else {
      int a = (int) car(n1);
      int b = (int) car(n2);
      int r = a % b;
      if (r < 0) r += b;
      return find_node(NUMBER, (node_ptr)r, Nil);
    }
  }
  
  if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_mod(WORD_NUMBER(car(n1)),
					      WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("mod", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Right shifts the first node by the number of bit given 
  in the second node.]

  Description        [First node should be NUMBER_WORD, and the second 
  one should be NUMBER]

  SideEffects        []

******************************************************************************/
node_ptr node_right_shift(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER) {
    WordNumber_ptr w = WORD_NUMBER(car(n1));
    int i = (int)car(n2);
    if ( i < 0 || i >= WordNumber_get_width(w)) {
      return failure_make("right operand of '>>' is out of range", 
			  FAILURE_UNSPECIFIED, yylineno);
    }
    return find_node(NUMBER_WORD, (node_ptr)WordNumber_right_shift(w, i), Nil);
  }
  else error_not_proper_numbers(">>", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Left shifts the first node by the number of bit given 
  in the second node.]

  Description        [First node should be NUMBER_WORD, and the second 
  one should be NUMBER]

  SideEffects        []

******************************************************************************/
node_ptr node_left_shift(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER) {
    WordNumber_ptr w = WORD_NUMBER(car(n1));
    int i = (int)car(n2);
    if ( i < 0 || i >= WordNumber_get_width(w)) {
      return failure_make("right operand of '<<' is out of range", 
			  FAILURE_UNSPECIFIED, yylineno);
    }
    return find_node(NUMBER_WORD, (node_ptr)WordNumber_left_shift(w, i), Nil);
  }
  else error_not_proper_numbers("<<", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Right rotates the first node by the number of bit given 
  in the second node.]

  Description        [First node should be NUMBER_WORD, and the second 
  one should be NUMBER]

  SideEffects        []

******************************************************************************/
node_ptr node_right_rotate(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */

  if (node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER) {
    WordNumber_ptr w = WORD_NUMBER(car(n1));
    int i = (int)car(n2);
    if ( i < 0 || i >= WordNumber_get_width(w)) {
      return failure_make("right operand of '>>>' is out of range", 
			  FAILURE_UNSPECIFIED, yylineno);
    }
    return find_node(NUMBER_WORD, (node_ptr)WordNumber_right_rotate(w, i), Nil);
  }
  else error_not_proper_numbers(">>>", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Left rotates the first node by the number of bit given 
  in the second node.]

  Description        [First node should be NUMBER_WORD, and the second 
  one should be NUMBER]

  SideEffects        []

******************************************************************************/
node_ptr node_left_rotate(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */
  
  if (node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER) {
    WordNumber_ptr w = WORD_NUMBER(car(n1));
    int i = (int)car(n2);
    if ( i < 0 || i >= WordNumber_get_width(w)) {
      return failure_make("right operand of '<<<' is out of range", 
			  FAILURE_UNSPECIFIED, yylineno);
    }
    return find_node(NUMBER_WORD, (node_ptr)WordNumber_left_rotate(w, i), Nil);
  }
  else error_not_proper_numbers("<<<", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [conatenates the values of two NUMBER_WORD nodes.]

  Description        []

  SideEffects        []

******************************************************************************/
node_ptr node_concatenation(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */
  
  /* cast */

  /* if n1 is boolean => cast to Word[1] */
  if (NUMBER == node_get_type(n1) && ((int)car(n1)&~1) == 0) {
    n1 = find_node(NUMBER_WORD, 
		   (node_ptr)WordNumber_from_integer((int)car(n1), 1), Nil);
  }

  /* if n2 is boolean => cast to Word[1] */
  if (NUMBER == node_get_type(n2) && ((int)car(n2)&~1) == 0) {
    n2 = find_node(NUMBER_WORD, 
		   (node_ptr)WordNumber_from_integer((int)car(n2), 1), Nil);
  }
  
  if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == NUMBER_WORD){
    return find_node(NUMBER_WORD,
	       (node_ptr)WordNumber_concatenate(WORD_NUMBER(car(n1)),
						WORD_NUMBER(car(n2))), Nil);
  }
  else error_not_proper_numbers("::", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [creates RANGE node from two NUMBER nodes.]

  Description        [this range is used in bit-selection only]

  SideEffects        [node_bit_selection]

******************************************************************************/
node_ptr node_bit_range(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */
  
  if (node_get_type(n1) == NUMBER && node_get_type(n2) == NUMBER) {
    return find_node(RANGE, n1, n2);
  }
  else error_not_proper_numbers("bit-selection-range", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [perform bit selection operation]

  Description        [first operand must be NUMBER_WORD node and 
  the second one must be RANGE(which holds the two NUMBERs).]

  SideEffects        [node_bits_range]

******************************************************************************/
node_ptr node_bit_selection(node_ptr n1, node_ptr n2)
{ 
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */
  
  /* cast */
  if(node_get_type(n1) == NUMBER_WORD && node_get_type(n2) == RANGE){
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_bit_select(WORD_NUMBER(car(n1)),
						     (int)car(car(n2)),
						     (int)car(cdr(n2))), Nil);
  }
  else error_not_proper_numbers("bit-selection", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [converts NUMBER_WORD node with 1-bit-width Word constant
  to boolean NUMBER.
 
  NOTE: At the momement, CUDD does not have unary 'apply', so
  you have to write a unary operator in the form of a binary one which
  actually applies to the first operand only]

  Description        []

  SideEffects        []

******************************************************************************/
node_ptr node_bool_cast(node_ptr n, node_ptr this_node_is_not_used)
{ 
  WordNumber_ptr word;

  if (node_get_type(n) == FAILURE) return n; /* error in previous expr */
  
  word = WORD_NUMBER(car(n));

  if (node_get_type(n) == NUMBER_WORD && WordNumber_get_width(word) == 1) {
    return  WordNumber_get_bit(word, 0) ? one_number : zero_number;
  }
  else error_not_proper_number("bool", n);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [ converts boolean NUMBER node to NUMBER_WORD
  with 1-bit-width Word constant.

  NOTE: At the momement, CUDD does not have unary 'apply', so
  you have to write a unary operator in the form of a binary one which
  actually applies to the first operand only]

  Description        []

  SideEffects        []

******************************************************************************/
node_ptr node_word1_cast(node_ptr n, node_ptr this_node_is_not_used)
{ 
  if (node_get_type(n) == FAILURE) return n; /* error in previous expr */
  
  if (node_get_type(n) == NUMBER) {
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_from_integer((int)car(n), 1), Nil);
  }
  else error_not_proper_number("word1", n);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Concatenates sign bit of n1 n2 number of times to n1]

  Description        [left exp must be a NUMBER_WORD and the right one
  is a NUMBER]

  SideEffects        []

******************************************************************************/
node_ptr node_sign_extend(node_ptr n1, node_ptr n2) {
  if (node_get_type(n1) == FAILURE) return n1; /* error in previous expr */
  if (node_get_type(n2) == FAILURE) return n2; /* error in previous expr */
  
  if (node_get_type(n1) == NUMBER_WORD && node_get_type(n2) ==NUMBER) {
    return find_node(NUMBER_WORD,
		     (node_ptr)WordNumber_sign_extend(WORD_NUMBER(car(n1)),
						      (int)car(n2)), Nil);
  }
  else error_not_proper_numbers("sign-extend", n1, n2);
  return (node_ptr)NULL;/* return something to suppress warnings */
}

/**Function********************************************************************

  Synopsis           [Computes the set union of two s_expr.]

  Description        [This function computes the sexp resulting from
  the union of s_expr "n1" and "n2".
  NB: if any of the operands is a FAILURE node, the FAILURE node is returned.]

  SideEffects        []

******************************************************************************/
node_ptr node_union(node_ptr n1, node_ptr n2)
{
  node_ptr tmp;

  if(n1 == Nil) return(n2);
  if(n2 == Nil) return(n1);

  if (node_get_type(n1) == FAILURE) return n1; /* error in operand */
  if (node_get_type(n2) == FAILURE) return n2; /* error in operand */
  
  /* convert singleton elements to lists */
  if(node_get_type(n1) != CONS) n1 = find_node(CONS, n1, Nil);
  if(node_get_type(n2) != CONS) n2 = find_node(CONS, n2, Nil);

  /* create a list merged from the given lists and 
     with all elements ordered (less-comparison)
  */
  tmp = Nil;
  while (n1 != Nil && n2 != Nil) {
    if (car(n1) == car(n2)) {
      tmp = cons(car(n1), tmp);
      n1 = cdr(n1);
      n2 = cdr(n2);
    }
    else if (car(n1) < car(n2)) {/* < is used because the list will be reversed */
      tmp = cons(car(n1), tmp);
      n1 = cdr(n1);
    }
    else { /*car(n2) > car(n1) */
      tmp = cons(car(n2), tmp);
      n2 = cdr(n2);
    }
  }
  if (Nil == n1) n1 = n2; /* remaining elements (they were in n1 or n2) */

  /* reverse the obtained list and apply find_node. The result will be in n1 */
  while (Nil != tmp) {
    n1 = find_node(CONS, car(tmp), n1);

    n2 =  cdr(tmp);
    free_node(tmp);
    tmp = n2;
  }
  return n1;
}

/**Function********************************************************************

  Synopsis           [Set inclusion]

  Description        [Checks if s_expr "n1" is a subset of s_expr
  "n2", if it is the case them <code>one_number</code> is returned,
  else <code>zero_number</code> is returned.

  If "n1" is a list of values then <code>one_number</code> is returned only
  if all elements of "n1" is a subset of "n2".

  NB: if any of the operands is a FAILURE node, the FAILURE node is returned.]

  SideEffects        []

  SeeAlso            []

******************************************************************************/
node_ptr node_setin(node_ptr n1, node_ptr n2)
{
  node_ptr iter1;
  node_ptr iter2;

  if (node_get_type(n1) == FAILURE) return n1; /* error in operand */
  if (node_get_type(n2) == FAILURE) return n2; /* error in operand */
  
  /* convert singleton elements to lists */
  if (CONS != node_get_type(n1)) n1 = find_node(CONS, n1, Nil);
  if (CONS != node_get_type(n2)) n2 = find_node(CONS, n2, Nil);
  
  /* check that every element of n1 is equal to some element of n2 */
  for (iter1 = n1; iter1 != Nil; iter1 = cdr(iter1)) {
    for (iter2 = n2; iter2 != Nil; iter2 = cdr(iter2)) {
      if (car(iter1) == car(iter2)) break; /* there is equality */
    }
    /* one of the elements of n1 is not equal to any elements of n2 */
    if (iter2 == Nil) return zero_number;
  }
  return one_number;
}

